/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class NestedLoopJoinExecutor : public AbstractExecutor {
private:
    std::unique_ptr<AbstractExecutor> left_;    // 左儿子节点（需要join的表）
    std::unique_ptr<AbstractExecutor> right_;   // 右儿子节点（需要join的表）
    size_t len_;                                // join后获得的每条记录的长度
    std::vector<ColMeta> cols_;                 // join后获得的记录的字段

    std::vector<Condition> fed_conds_;          // join条件
    std::unique_ptr<RmRecord> result;           // 存储当前迭代轮次的值，供`Next`取走
    std::vector<std::unique_ptr<RmRecord>> left_record;     
    std::vector<std::unique_ptr<RmRecord>> right_record;    
    std::vector<std::unique_ptr<RmRecord>>::iterator left_itrator;
    std::vector<std::unique_ptr<RmRecord>>::iterator right_itrator;   
    bool isend;

public:
    NestedLoopJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right,
                           std::vector<Condition> conds) {
        left_ = std::move(left);
        right_ = std::move(right);
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        auto right_cols = right_->cols();
        for (auto &col: right_cols) {
            col.offset += left_->tupleLen();
        }
        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        isend = false;
        fed_conds_ = std::move(conds);
    }

    void move_next() {
        right_itrator++;
        if (right_itrator == right_record.end()) {
            left_itrator++;
            right_itrator = right_record.begin();
            if (left_itrator == left_record.end()) {
                isend = true; 
                return;
            }
        }
    }

    /**
     * Initializes the iterator to the first tuple.
     * Iterates over all tuples in the left and right child executors,
     * and stores them in `left_record` and `right_record`.
     * Sets `lit` and `rit` to the beginning of `left_record` and `right_record`,
     * respectively. Sets `isend` to true if either `left_record` or `right_record` is empty.
     * Steps through the join conditions until either `isend` is true or a tuple satisfies the join conditions.
     * If `isend` is false, creates a new `RmRecord` with the joined tuple data and sets `result` to it.
     */
    void beginTuple() override {
        // Iterate over all tuples in the left child executor
        for (left_->beginTuple(); !left_->is_end(); left_->nextTuple()) {
            left_record.push_back(left_->Next());
        }

        // Iterate over all tuples in the right child executor
        for (right_->beginTuple(); !right_->is_end(); right_->nextTuple()) {
            right_record.push_back(right_->Next());
        }

        // Set `lit` and `rit` to the beginning of `left_record` and `right_record`
        left_itrator = left_record.begin();
        right_itrator = right_record.begin();

        // Set `isend` to true if either `left_record` or `right_record` is empty
        isend = left_itrator == left_record.end() || right_itrator == right_record.end();

        // Step through the join conditions until either `isend` is true or a tuple satisfies the join conditions
        while (!isend && !evalConditions()) {
            move_next();
        }

        // If `isend` is false, create a new `RmRecord` with the joined tuple data and set `result` to it
        if (!isend) {
            result = std::make_unique<RmRecord>(len_);
            memcpy(result->data, (*left_itrator)->data, left_->tupleLen());
            memcpy(result->data + left_->tupleLen(), (*right_itrator)->data, right_->tupleLen());
        }
    }

    /**
     * Move the iterator to the next tuple.
     * It first steps through the join conditions until either `isend` is true or a tuple satisfies the join conditions.
     * If `isend` is false, creates a new `RmRecord` with the joined tuple data and sets `result` to it.
     *
     * @throws std::logic_error if the iterator is already at the end.
     */
    void nextTuple() override {
        // Check if iterator is at the end
        assert(!is_end());

        // Step through the join conditions until either `isend` is true or a tuple satisfies the join conditions
        move_next();
        while (!isend && !evalConditions()){
            move_next();
        }

        // If iterator is still not at the end, create a new `RmRecord` with the joined tuple data and set `result` to it
        if (isend) {
            return;
        }

        assert(result == nullptr);

        // Create a new `RmRecord` with the joined tuple data
        result = std::make_unique<RmRecord>(len_);

        // Copy the left tuple data to the new `RmRecord`
        memcpy(result->data, (*left_itrator)->data, left_->tupleLen());

        // Copy the right tuple data to the new `RmRecord`
        memcpy(result->data + left_->tupleLen(), (*right_itrator)->data, right_->tupleLen());

    }


    [[nodiscard]] bool is_end() const override {
        return isend;
    }

    [[nodiscard]] size_t tupleLen() const override {
        return len_;
    };

    bool evalConditions() {
        char *lbase = (*left_itrator)->data;
        char *rbase = (*right_itrator)->data;
        return std::all_of(fed_conds_.begin(), fed_conds_.end(), [&](Condition &cond) {
            assert(!cond.is_rhs_val);
            auto lvalue = convertValue(lbase, get_col_offset_lr(left_->cols(), cond.lhs_col));
            auto rvalue = convertValue(rbase, get_col_offset_lr(right_->cols(), cond.rhs_col));
            return cond.eval(lvalue, rvalue);
        });
    }

    static ColMeta get_col_offset_lr(const std::vector<ColMeta> &cols, const TabCol &target) {
        auto it = std::find_if(cols.begin(), cols.end(), [&target](const ColMeta &col) {
            return col.tab_name == target.tab_name && col.name == target.col_name;
        });
        assert(it != cols.end());
        return *it;
    }

    [[nodiscard]] const std::vector<ColMeta> &cols() const override {
        return cols_;
    }

    std::unique_ptr<RmRecord> Next() override {
        return std::move(result);
    }

    Rid &rid() override { return _abstract_rid; }
};