/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
#include <memory>
#include <vector>

class UpdateExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;
    std::vector<Condition> conds_;
    RmFileHandle *fh_;
    std::vector<Rid> rids_;
    std::string tab_name_;
    std::vector<SetClause> set_clauses_;
    SmManager *sm_manager_;

   public:
    UpdateExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<SetClause> set_clauses,
                   std::vector<Condition> conds, std::vector<Rid> rids, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        set_clauses_ = set_clauses;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;
        context_ = context;
    }
    std::unique_ptr<RmRecord> Next() override {
        int record_size = fh_->get_file_hdr().record_size;
        auto buf = std::make_unique<char[]>(record_size);
        for (const auto &rid: rids_) {
            std::unique_ptr<RmRecord> record = fh_->get_record(rid, context_);
            memcpy(buf.get(), record->data, record_size);
            for (auto &clause: set_clauses_) {
                std::vector<ColMeta>::iterator col = tab_.get_col(clause.lhs.col_name);
                clause.rhs.init_raw(col->len);
                memcpy(buf.get() + col->offset, clause.rhs.raw->data, col->len);
            }

            RmRecord old_record{record->size};
            RmRecord new_record{record->size};

            memcpy(old_record.data, record->data, record->size);
            memcpy(new_record.data, buf.get(), record->size);
            // 检查更新主键
            for (size_t i = 0; i < tab_.indexes.size(); ++i) {
                auto &index = tab_.indexes[i];
                auto ih = sm_manager_->ihs_.at(index.index_name).get();
                auto old_key = index.get_key(old_record.data);
                auto insert_key = index.get_key(new_record.data);
                if (memcmp(old_key.get(), insert_key.get(), index.col_tot_len) == 0) {
                    continue;
                }
                std::vector<Rid> result;
                if (ih->get_value(insert_key.get(), &result, context_->txn_)) {
                    throw DuplicateKeyError(index.index_name);
                }
            }
            
            // 更新record
            fh_->update_record(rid, buf.get(), context_);

            // 在 index 中更新
            for(size_t i = 0; i < tab_.indexes.size(); ++i) {
                auto& index = tab_.indexes[i];
                auto const &cols = index.cols;
                // 需要 update 的字段是否在索引中出现
                bool need_update = false;
                for (auto const &set_clause : set_clauses_) {
                    auto pos = std::find_if(cols.begin(), cols.end(), [&set_clause](const ColMeta &col) {
                        return col.name == set_clause.lhs.col_name;
                    });
                    if (pos != cols.end()) {
                        need_update = true;
                        break;
                    }
                }
                if (!need_update) {
                    continue;
                }
                auto ih = sm_manager_->ihs_.at(index.index_name).get();
                auto delete_key = index.get_key(old_record.data);
                auto insert_key = index.get_key(new_record.data);
                ih->delete_entry(delete_key.get(), context_->txn_);
                ih->insert_entry(insert_key.get(), rid, context_->txn_);
            }
        }
        return nullptr;
    }

    Rid &rid() override { return _abstract_rid; }
};