/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "ix_index_handle.h"

#include "ix_scan.h"

int IxNodeHandle::lower_bound(const char *target) const {
    int low = 0, high = page_hdr->num_key;
    while (low < high) {
        int mid = (low + high) / 2;
        int comparison = ix_compare(get_key(mid), target, file_hdr->col_types_, file_hdr->col_lens_);
        if (comparison < 0) {
            low = mid + 1;
        } else {
            high = mid;
        }
    }
    return low;
}

/**
 * @brief Find the index of the first key in the current node that is greater than target.
 *
 * @return The key index, in the range [1, num_key), if the returned key index is equal to num_key, it means target is greater than or equal to the last key.
 * @note The range starts from 1.
 */
int IxNodeHandle::upper_bound(const char *target) const {
    if (binary_search) {
        int left = 1, right = page_hdr->num_key;
        while (left < right) {
            int mid = (left + right) >> 1;
            if (ix_compare(get_key(mid), target, file_hdr->col_types_, file_hdr->col_lens_) <= 0)
                left = mid + 1;
            else
                right = mid;
        }
        return left;
    }
    for (int i = 1; i < page_hdr->num_key; i++) {
        if (ix_compare(get_key(i), target, file_hdr->col_types_, file_hdr->col_lens_) > 0) {
            return i;
        }
    }
    return page_hdr->num_key;
}

bool IxNodeHandle::leaf_lookup(const char *targetKey, Rid **targetValue) {
    int position = lower_bound(targetKey);
    if (position < page_hdr->num_key &&
        ix_compare(get_key(position), targetKey, file_hdr->col_types_, file_hdr->col_lens_) == 0) {
        *targetValue = get_rid(position);
        return true;
    }
    return false;
}

/**
 * 用于内部结点（非叶子节点）查找目标key所在的孩子结点（子树）
 * @param key 目标key
 * @return page_id_t 目标key所在的孩子节点（子树）的存储页面编号
 */
page_id_t IxNodeHandle::internal_lookup(const char *key) {
    // Todo:
    // 1. 查找当前非叶子节点中目标key所在孩子节点（子树）的位置
    // 2. 获取该孩子节点（子树）所在页面的编号
    // 3. 返回页面编号

    int pos = upper_bound(key) - 1;
    return get_rid(pos)->page_no;
}

/**
 * @brief 在指定位置插入n个连续的键值对
 * 将key的前n位插入到原来keys中的pos位置；将rid的前n位插入到原来rids中的pos位置
 *
 * @param pos 要插入键值对的位置
 * @param (key, rid) 连续键值对的起始地址，也就是第一个键值对，可以通过(key, rid)来获取n个键值对
 * @param n 键值对数量
 * @note [0,pos)           [pos,num_key)
 *                            key_slot
 *                            /      \
 *                           /        \
 *       [0,pos)     [pos,pos+n)   [pos+n,num_key+n)
 *                      key           key_slot
 */
void IxNodeHandle::insert_pairs(int pos, const char *key, const Rid *rid, int n) {
    // Todo:
    // 1. 判断pos的合法性
    // 2. 通过key获取n个连续键值对的key值，并把n个key值插入到pos位置
    // 3. 通过rid获取n个连续键值对的rid值，并把n个rid值插入到pos位置
    // 4. 更新当前节点的键数量

    if (pos < 0 && pos > page_hdr->num_key) {
        return;
    }

    if (pos < page_hdr->num_key) {
        int move_num = page_hdr->num_key - pos;
        memmove(keys + (pos + n) * file_hdr->col_tot_len_, keys + pos * file_hdr->col_tot_len_,
                move_num * file_hdr->col_tot_len_);
        memmove(rids + pos + n, rids + pos, move_num * sizeof(Rid));
    }
    memmove(keys + pos * file_hdr->col_tot_len_, key, n * file_hdr->col_tot_len_);
    memmove(rids + pos, rid, n * sizeof(Rid));

    page_hdr->num_key += n;
}

/**
 * @brief 用于在结点中插入单个键值对。
 * 函数返回插入后的键值对数量
 *
 * @param (key, value) 要插入的键值对
 * @return int 键值对数量
 */
int IxNodeHandle::insert(const char *key, const Rid &value) {
    // Todo:
    // 1. 查找要插入的键值对应该插入到当前节点的哪个位置
    // 2. 如果key重复则不插入
    // 3. 如果key不重复则插入键值对
    // 4. 返回完成插入操作之后的键值对数量

    // 插入的位置应该是第一个大于等于key的位置
    int pos = lower_bound(key);
    // 如果这个位置没有key，那么直接插入
    if (pos >= page_hdr->num_key) {
        pos = page_hdr->num_key;
        insert_pair(pos, key, value);
        return page_hdr->num_key;
    }
    // 如果这个位置有key，那么判断是否重复
    if (ix_compare(get_key(pos), key, file_hdr->col_types_, file_hdr->col_lens_) != 0) {
        insert_pair(pos, key, value);
    }
    return page_hdr->num_key;
}

/**
 * @brief 用于在结点中的指定位置删除单个键值对
 *
 * @param pos 要删除键值对的位置
 */
void IxNodeHandle::erase_pair(int pos) {
    // Todo:
    // 1. 删除该位置的key
    // 2. 删除该位置的rid
    // 3. 更新结点的键值对数量
    if (pos < 0 && pos >= get_size()) {
        return;
    }

    // 将 pos 后面的全部前移一位
    char *key_begin = get_key(pos);
    Rid *rid_begin = get_rid(pos);
    // 需要前移的键值对个数
    int num = get_size() - pos - 1;
    memmove(key_begin, key_begin + file_hdr->col_tot_len_, num * file_hdr->col_tot_len_);
    memmove(rid_begin, rid_begin + 1, num * sizeof(Rid));
    page_hdr->num_key--;
}

/**
 * @brief 用于在结点中删除指定key的键值对。函数返回删除后的键值对数量
 *
 * @param key 要删除的键值对key值
 * @return 完成删除操作后的键值对数量
 */
int IxNodeHandle::remove(const char *key) {
    // Todo:
    // 1. 查找要删除键值对的位置
    // 2. 如果要删除的键值对存在，删除键值对
    // 3. 返回完成删除操作后的键值对数量

    int pos = lower_bound(key);
    if (pos < page_hdr->num_key && ix_compare(get_key(pos), key, file_hdr->col_types_, file_hdr->col_lens_) == 0) {
        erase_pair(pos);
    }
    return page_hdr->num_key;
}

IxIndexHandle::IxIndexHandle(DiskManager *disk_manager, BufferPoolManager *buffer_pool_manager, int fd)
    : disk_manager_(disk_manager), buffer_pool_manager_(buffer_pool_manager), fd_(fd) {
    // init file_hdr_
    disk_manager_->read_page(fd, IX_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
    char *buf = new char[PAGE_SIZE];
    memset(buf, 0, PAGE_SIZE);
    disk_manager_->read_page(fd, IX_FILE_HDR_PAGE, buf, PAGE_SIZE);
    file_hdr_ = new IxFileHdr();
    file_hdr_->deserialize(buf);

    // disk_manager管理的fd对应的文件中，设置从file_hdr_->num_pages开始分配page_no
    int now_page_no = disk_manager_->get_fd2pageno(fd);
    disk_manager_->set_fd2pageno(fd, now_page_no + 1);
    delete[] buf;
}

/**
 * @brief Find the leaf page that contains the specified key.
 *
 * @param key The target key value.
 * @param operation The operation type to perform after finding the target key value.
 * @param transaction The transaction pointer. If not needed, pass nullptr.
 * @return [leaf node] and [root_is_latched] The target leaf node and whether the root node is locked.
 */
std::pair<IxNodeHandle *, bool> IxIndexHandle::find_leaf_page(const char *key, Operation operation,
                                                              Transaction *transaction, bool find_first) {
    // Get the root node
    root_latch_.lock();
    bool is_root_locked = true;
    page_id_t root_page = file_hdr_->root_page_;
    if (root_page == INVALID_PAGE_ID) {
        unlock_pages(transaction, operation, is_root_locked);
        return {nullptr, is_root_locked};
    }
    IxNodeHandle *root_node = fetch_node(root_page);
    if (operation == Operation::FIND) {
        root_latch_.unlock();
        is_root_locked = false;
    } else {
        if (is_node_safe(root_node, operation)) {
            unlock_pages(transaction, operation, is_root_locked);
        }
    }

    // Traverse down the B+ tree to find the leaf page that contains the key
    IxNodeHandle *node = root_node;
    while (!node->is_leaf_page()) {
        page_id_t child_page = node->internal_lookup(key);
        IxNodeHandle *child_node = fetch_node(child_page);
        if (operation != Operation::FIND) {
            if (transaction)
                transaction->append_index_latch_page_set(node->page);
            if (!is_node_safe(child_node, operation)) {
                unlock_pages(transaction, operation, is_root_locked);
            }
        }
        delete node;
        node = child_node;
    }
    if (operation != Operation::FIND) {
        if (transaction)
            transaction->append_index_latch_page_set(node->page);
    }

    // Return the leaf node
    return {node, is_root_locked};
}

/**
 * @brief 用于查找指定键在叶子结点中的对应的值result，B+ 树查找入口
 *
 * @param key 查找的目标key值
 * @param result 用于存放结果的容器
 * @param transaction 事务指针
 * @return bool 返回目标键值对是否存在
 */
bool IxIndexHandle::get_value(const char *key, std::vector<Rid> *result, Transaction *transaction) {
    // Todo:
    // 1. 获取目标key值所在的叶子结点
    // 2. 在叶子节点中查找目标key值的位置，并读取key对应的rid
    // 3. 把rid存入result参数中
    // 提示：使用完buffer_pool提供的page之后，记得unpin page；记得处理并发的上锁

    auto pair = find_leaf_page(key, Operation::FIND, transaction);
    IxNodeHandle *node = pair.first;
    Rid *value = nullptr;
    bool found = false;
    if (node) {
        if (node->leaf_lookup(key, &value)) {
            result->push_back(*value);
            found = true;
        }
    }
    unlock_pages(transaction, Operation::FIND, pair.second);
    if (node) {
        delete node;
    }

    return found;
}

/**
 * @brief  将传入的一个node拆分(split)成两个结点，在node的右边生成一个新结点new node
 * @param node 需要拆分的结点
 * @return 拆分得到的new_node
 * @note need to unpin the new node outside
 * 注意：本函数执行完毕后，原node和new node都需要在函数外面进行unpin
 */
IxNodeHandle *IxIndexHandle::split(IxNodeHandle *node) {
    // Todo:
    // 1. 将原结点的键值对平均分配，右半部分分裂为新的右兄弟结点
    //    需要初始化新节点的page_hdr内容
    // 2. 如果新的右兄弟结点是叶子结点，更新新旧节点的prev_leaf和next_leaf指针
    //    为新节点分配键值对，更新旧节点的键值对数记录
    // 3. 如果新的右兄弟结点不是叶子结点，更新该结点的所有孩子结点的父节点信息(使用IxIndexHandle::maintain_child())

    assert(node->get_size() == node->get_max_size());

    IxNodeHandle *new_node = create_node();
    // 初始化page_hdr 内容
    if ((new_node->page_hdr->is_leaf = node->is_leaf_page()) == true) {
        new_node->set_prev_leaf(node->get_page_no());
        new_node->set_next_leaf(node->get_next_leaf());
        node->set_next_leaf(new_node->get_page_no());
        IxNodeHandle *next_node = fetch_node(new_node->page_hdr->next_leaf);
        next_node->page_hdr->prev_leaf = new_node->get_page_no();
        // buffer_pool_manager_->unpin_page(next_node->get_page_id(), true);
        delete next_node;
    }
    new_node->page_hdr->num_key = 0;
    new_node->page_hdr->next_free_page_no = node->page_hdr->next_free_page_no;
    new_node->set_parent_page_no(node->get_parent_page_no());

    // 将原node中[minsize, maxsize)的节点移动到新的node中
    int pos = node->get_min_size();
    int n = node->get_max_size() - pos;
    // new_node 增加 num_key 在 insert_pairs 函数里面实现
    new_node->insert_pairs(0, node->get_key(pos), node->get_rid(pos), n);
    node->page_hdr->num_key -= n;

    // 如果不是叶子节点，则更新其所有孩子节点的父节点信息
    if (!new_node->is_leaf_page()) {
        for (int i = 0; i < n; i++) maintain_child(new_node, i);
    }

    return new_node;
}

/**
 * @brief Insert key & value pair into internal page after split
 * 拆分(split)后，向上找到old_node的父结点
 * 将new_node的第一个key插入到父结点，其位置在 父结点指向old_node的孩子指针 之后
 * 如果插入后>=maxsize，则必须继续拆分父结点，然后在其父结点的父结点再插入，即需要递归
 * 直到找到的old_node为根结点时，结束递归（此时将会新建一个根R，关键字为key，old_node和new_node为其孩子）
 *
 * @param (old_node, new_node) 原结点为old_node，old_node被分裂之后产生了新的右兄弟结点new_node
 * @param key 要插入parent的key
 * @note 一个结点插入了键值对之后需要分裂，分裂后左半部分的键值对保留在原结点，在参数中称为old_node，
 * 右半部分的键值对分裂为新的右兄弟节点，在参数中称为new_node（参考split函数来理解old_node和new_node）
 * @note 本函数执行完毕后，new node和old node都需要在函数外面进行unpin
 */
void IxIndexHandle::insert_into_parent(IxNodeHandle *old_node, const char *key, IxNodeHandle *new_node,
                                       Transaction *transaction) {
    // Todo:
    // 1. 分裂前的结点（原结点, old_node）是否为根结点，如果为根结点需要分配新的root
    // 2. 获取原结点（old_node）的父亲结点
    // 3. 获取key对应的rid，并将(key, rid)插入到父亲结点
    // 4. 如果父亲结点仍需要继续分裂，则进行递归插入
    // 提示：记得unpin page

    // 如果是根节点，则需要新建一个根节点，其子节点为old_node 与 new_node
    if (old_node->is_root_page()) {
        // 创建新的根结点
        IxNodeHandle *new_root = create_node();
        new_root->page_hdr->is_leaf = 0; // 新根结点不是叶子结点
        new_root->page_hdr->num_key = 0; // 新根结点没有键值对
        new_root->page_hdr->parent = INVALID_PAGE_ID; // 新根结点没有父结点
        new_root->page_hdr->next_free_page_no = IX_NO_PAGE; // 新根结点没有兄弟结点

        // 将原结点的第一个键值对插入到新根结点
        new_root->insert_pair(0, old_node->get_key(0), (Rid){old_node->get_page_no(), -1});
        // 将新的键值对插入到新根结点
        new_root->insert_pair(1, key, (Rid){new_node->get_page_no(), -1});

        // 更新根结点的page_no
        int new_root_page = new_root->get_page_no();
        file_hdr_->root_page_ = new_root_page;
        // 更新原结点与新结点的父结点信息
        old_node->set_parent_page_no(new_root_page);
        new_node->set_parent_page_no(new_root_page);
        // 释放新根结点占用的缓冲区
        // buffer_pool_manager_->unpin_page(new_root->get_page_id(), true);
        delete new_root;
        return;
    }
    // 如果不是根节点，则需要获取原结点的父结点，将新的键值对插入其中
    IxNodeHandle *parent_node = fetch_node(old_node->get_parent_page_no());
    int new_node_pos = parent_node->find_child(old_node) + 1; // 新结点插入的位置
    parent_node->insert_pair(new_node_pos, key, (Rid){new_node->get_page_no(), -1}); // 将新的键值对插入到父结点
    // 如果插入后父结点满了，则需要继续分裂
    if (parent_node->get_size() == parent_node->get_max_size()) {
        IxNodeHandle *parent_bro_node = split(parent_node); // 分裂父结点
        insert_into_parent(parent_node, parent_bro_node->get_key(0), parent_bro_node, transaction); // 递归插入到父结点的父结点
        // 释放分裂后的新结点占用的缓冲区
        // buffer_pool_manager_->unpin_page(parent_bro_node->get_page_id(), true);
        delete parent_bro_node;
    }
    // 释放原结点的父结点占用的缓冲区
    // buffer_pool_manager_->unpin_page(parent_node->get_page_id(), true);
    delete parent_node;
}

/**
 * @brief 将指定键值对插入到B+树中
 * @param (key, value) 要插入的键值对
 * @param transaction 事务指针
 * @return page_id_t 插入到的叶结点的page_no
 */
page_id_t IxIndexHandle::insert_entry(const char *key, const Rid &value, Transaction *transaction) {
    // Todo:
    // 1. 查找key值应该插入到哪个叶子节点
    // 2. 在该叶子节点中插入键值对
    // 3. 如果结点已满，分裂结点，并把新结点的相关信息插入父节点
    // 提示：记得unpin page；若当前叶子节点是最右叶子节点，则需要更新file_hdr_.last_leaf；记得处理并发的上锁

    // Find the leaf page to insert the key-value pair into
    auto pair = find_leaf_page(key, Operation::INSERT, transaction);
    IxNodeHandle *node = pair.first;

    // Record the size before insertion
    int pre_insert_sz = node->get_size();

    // Insert the key-value pair into the leaf node
    int after_insert_sz = node->insert(key, value);

    // If the size remains unchanged after insertion, it means the insertion failed (duplicate key)
    if (pre_insert_sz == after_insert_sz) {
        // Unlock the pages and delete the node
        unlock_pages(transaction, Operation::INSERT, pair.second);
        delete node;
        return INVALID_PAGE_ID;
    }

    // If the node is full after insertion, split it and update the parent node information
    if (node->get_size() == node->get_max_size()) {
        // Split the node and get the new node
        IxNodeHandle *new_node = split(node);

        // Update the file header if the original node is the last leaf
        if (node->get_page_no() == file_hdr_->last_leaf_) {
            file_hdr_->last_leaf_ = new_node->get_page_no();
        }

        // Insert the key-value pair into the parent node and update the parent node information
        insert_into_parent(node, new_node->get_key(0), new_node, transaction);

        // Delete the new node
        delete new_node;
    }

    // Get the page number of the inserted leaf node
    page_id_t page_no = node->get_page_id().page_no;

    // Unlock the pages and delete the leaf node
    unlock_pages(transaction, Operation::INSERT, pair.second);
    delete node;

    return page_no;
}

/**
 * @brief Deletes a key-value pair from the B+ tree index.
 *
 * @param key The key to be deleted.
 * @param transaction The transaction pointer.
 * @return true if the deletion is successful, false otherwise.
 */
bool IxIndexHandle::delete_entry(const char *key, Transaction *transaction) {
    auto leaf_page = find_leaf_page(key, Operation::DELETE, transaction);
    IxNodeHandle *leaf = leaf_page.first;
    int prev_size = leaf->get_size();

    if (prev_size == leaf->remove(key)) {
        unlock_pages(transaction, Operation::DELETE, leaf_page.second);
        delete leaf;
        return false;
    }

    if (coalesce_or_redistribute(leaf, transaction)) {
        delete_pages(transaction);
    }

    unlock_pages(transaction, Operation::DELETE, leaf_page.second);
    delete leaf;

    return true;
}

/**
 * @brief 用于处理合并和重分配的逻辑，用于删除键值对后调用
 *
 * @param node 执行完删除操作的结点
 * @param transaction 事务指针
 * @param root_is_latched 传出参数：根节点是否上锁，用于并发操作
 * @return 是否需要删除结点
 * @note User needs to first find the sibling of input page.
 * If sibling's size + input page's size >= 2 * page's minsize, then redistribute.
 * Otherwise, merge(coalesce).
 */
bool IxIndexHandle::coalesce_or_redistribute(IxNodeHandle *node, Transaction *transaction, bool *root_is_latched) {
    // Todo:
    // 1. 判断node结点是否为根节点
    //    1.1 如果是根节点，需要调用adjust_root() 函数来进行处理，返回根节点是否需要被删除
    //    1.2 如果不是根节点，并且不需要执行合并或重分配操作，则直接返回false，否则执行2
    // 2. 获取node结点的父亲结点
    // 3. 寻找node结点的兄弟结点（优先选取前驱结点）
    // 4. 如果node结点和兄弟结点的键值对数量之和，能够支撑两个B+树结点（即node.size+neighbor.size >=
    // NodeMinSize*2)，则只需要重新分配键值对（调用redistribute函数）
    // 5. 如果不满足上述条件，则需要合并两个结点，将右边的结点合并到左边的结点（调用coalesce函数）

    // 判断是否为根节点，如果是根节点，调用adjust_root()函数处理
    if (node->is_root_page()) {
        // 如果根节点删除了一个键值对，需要判断是否需要删除根节点
        if (adjust_root(node)) {
            // 如果事务不为空，则将根节点添加到事务的latch page set中
            if (transaction) transaction->append_index_latch_page_set(node->page);
            return true;
        }
        return false;
    }
    // 如果不是根节点，并且不需要执行合并或重分配操作
    if (node->get_size() >= node->get_min_size()) {
        // 调整父节点的信息
        maintain_parent(node);
        return false;
    }
    // 是否要删除节点
    bool is_delete = false;
    // 获取父节点
    IxNodeHandle *parent_node = fetch_node(node->get_parent_page_no());
    // 获取兄弟节点
    IxNodeHandle *bro_node = nullptr;
    int pos = parent_node->find_child(node);
    if (pos != 0) {
        // 如果有左兄弟，选择左兄弟
        bro_node = fetch_node(parent_node->value_at(pos - 1));
    } else {
        // 否则选择右兄弟
        bro_node = fetch_node(parent_node->value_at(pos + 1));
    }
    // 如果事务不为空，则将兄弟节点添加到事务的latch page set中
    if(transaction)
        transaction->append_index_latch_page_set(bro_node->page);
    // 判断是否满足重新分配条件
    if (node->get_size() + bro_node->get_size() >= node->get_min_size() * 2) {
        // 如果满足重新分配条件，调用redistribute()函数进行重新分配
        redistribute(bro_node, node, parent_node, pos);
    } else {
        // 如果不满足重新分配条件，则调用coalesce()函数进行合并
        if (coalesce(&bro_node, &node, &parent_node, pos, transaction, nullptr)) {
            // 如果事务不为空，则将父节点添加到事务的latch page set中
            if(transaction)
                transaction->append_index_latch_page_set(parent_node->page);
        }
        is_delete = true;
    }
    // 释放兄弟节点和父节点的页面
    delete bro_node;
    delete parent_node;
    return is_delete;
}

/**
 * @brief 用于当根结点被删除了一个键值对之后的处理
 * @param old_root_node 原根节点
 * @return bool 根结点是否需要被删除
 * @note size of root page can be less than min size and this method is only called within coalesce_or_redistribute()
 */
bool IxIndexHandle::adjust_root(IxNodeHandle *old_root_node) {
    // Todo:
    // 1. 如果old_root_node是内部结点，并且大小为1，则直接把它的孩子更新成新的根结点
    // 2. 如果old_root_node是叶结点，且大小为0，则直接更新root page
    // 3. 除了上述两种情况，不需要进行操作

    // 如果原根节点是内部节点，且大小为1，说明B+树只有一个节点，将其唯一子节点作为新的根节点
    // 1. 获取原根节点的唯一子节点
    // 2. 更新根页面为新子节点的页面号
    // 3. 将新子节点的父节点设为无效页面号
    // 4. 释放旧根节点的页面
    // 5. 返回true，表示根节点需要被删除
    if (!old_root_node->is_leaf_page() && old_root_node->get_size() == 1) {
        IxNodeHandle *node = fetch_node(old_root_node->value_at(0));
        file_hdr_->root_page_ = node->get_page_no();
        node->set_parent_page_no(INVALID_PAGE_ID);
        release_node_handle(*old_root_node);
        delete node;
        return true;
    }

    return false;
}

/**
 * @brief 重新分配node和兄弟结点neighbor_node的键值对
 * redistribute key & value pairs from one page to its sibling page. If index == 0, move sibling page's first key
 * & value pair into end of input "node", otherwise move sibling page's last key & value pair into head of input "node".
 *
 * @param neighbor_node sibling page of input "node"
 * @param node input from method coalesce_or_redistribute()
 * @param parent the parent of "node" and "neighbor_node"
 * @param index node在parent中的rid_idx
 * @note node是之前刚被删除过一个key的结点
 * index=0，则neighbor是node后继结点，表示：node(left)      neighbor(right)
 * index>0，则neighbor是node前驱结点，表示：neighbor(left)  node(right)
 * 注意更新parent结点的相关kv对
 */
void IxIndexHandle::redistribute(IxNodeHandle *neighbor_node, IxNodeHandle *node, IxNodeHandle *parent, int index) {
    // Todo:
    // 1. 通过index判断neighbor_node是否为node的前驱结点
    // 2. 从neighbor_node中移动一个键值对到node结点中
    // 3. 更新父节点中的相关信息，并且修改移动键值对对应孩字结点的父结点信息（maintain_child函数）
    // 注意：neighbor_node的位置不同，需要移动的键值对不同，需要分类讨论

    if (index == 0) {
        node->insert_pair(node->get_size(), neighbor_node->get_key(0), *neighbor_node->get_rid(0));
        neighbor_node->erase_pair(0);
        maintain_child(node, node->get_size() - 1);
        maintain_parent(neighbor_node);
    } else {
        int pos = neighbor_node->get_size() - 1;
        node->insert_pair(0, neighbor_node->get_key(pos), *neighbor_node->get_rid(pos));
        neighbor_node->erase_pair(pos);
        maintain_child(node, 0);
        maintain_parent(node);
    }
}

/**
 * @brief 合并(coalesce)函数是将node和其直接前驱进行合并，也就是和它左边的neighbor_node进行合并；
 * 假设node一定在右边。如果上层传入的index=0，说明node在左边，那么交换node和neighbor_node，保证node在右边；合并到左结点，实际上就是删除了右结点；
 * Move all the key & value pairs from one page to its sibling page, and notify buffer pool manager to delete this page.
 * Parent page must be adjusted to take info of deletion into account. Remember to deal with coalesce or redistribute
 * recursively if necessary.
 *
 * @param neighbor_node sibling page of input "node" (neighbor_node是node的前结点)
 * @param node input from method coalesce_or_redistribute() (node结点是需要被删除的)
 * @param parent parent page of input "node"
 * @param index node在parent中的rid_idx
 * @return true means parent node should be deleted, false means no deletion happend
 * @note Assume that *neighbor_node is the left sibling of *node (neighbor -> node)
 */
bool IxIndexHandle::coalesce(IxNodeHandle **neighbor_node, IxNodeHandle **node, IxNodeHandle **parent, int index,
                             Transaction *transaction, bool *root_is_latched) {
    // Todo:
    // 1. 用index判断neighbor_node是否为node的前驱结点，若不是则交换两个结点，让neighbor_node作为左结点，node作为右结点
    // 2. 把node结点的键值对移动到neighbor_node中，并更新node结点孩子结点的父节点信息（调用maintain_child函数）
    // 3. 释放和删除node结点，并删除parent中node结点的信息，返回parent是否需要被删除
    // 提示：如果是叶子结点且为最右叶子结点，需要更新file_hdr_.last_leaf

    // 交换 node 和 neighbor_node 的位置，保证 node 为右结点
    IxNodeHandle *left_node = (index == 0) ? *node : *neighbor_node;
    IxNodeHandle *right_node = (index == 0) ? *neighbor_node : *node;
    // 更新 index 为 right_node 在父节点中的位置
    index = (index == 0) ? 1 : index;

    // 将 right_node 的键值对移动到 left_node 中
    int pre_sz = left_node->get_size();
    left_node->insert_pairs(pre_sz, right_node->get_key(0), right_node->get_rid(0), right_node->get_size());
    for (int i = pre_sz; i < left_node->get_size(); i++) {
        maintain_child(left_node, i);
    }

    // 如果 right_node 是最右叶子结点，需要更新 file_hdr_->last_leaf_
    if (right_node->get_page_no() == file_hdr_->last_leaf_) {
        file_hdr_->last_leaf_ = left_node->get_page_no();
    }
    // 在事务中添加 right_node 的 latch 页
    if (transaction) transaction->append_index_latch_page_set(right_node->page);
    // 在 erase_leaf 中修改 right_node 的 prev_leaf_ 和 next_leaf_ 信息
    if (right_node->is_leaf_page()) erase_leaf(right_node);
    // 释放 right_node 的内存
    release_node_handle(*right_node);
    // 在父节点中删除 node 相关键值对
    (*parent)->erase_pair(index);
    // 继续 coalesce_or_redistribute 或重新分配
    return coalesce_or_redistribute(*parent, transaction);
}

/**
 * @brief 这里把iid转换成了rid，即iid的slot_no作为node的rid_idx(key_idx)
 * node其实就是把slot_no作为键值对数组的下标
 * 换而言之，每个iid对应的索引槽存了一对(key,rid)，指向了(要建立索引的属性首地址,插入/删除记录的位置)
 *
 * @param iid
 * @return Rid
 * @note iid和rid存的不是一个东西，rid是上层传过来的记录位置，iid是索引内部生成的索引槽位置
 */
Rid IxIndexHandle::get_rid(const Iid &iid) const {
    IxNodeHandle *node = fetch_node(iid.page_no);
    if (iid.slot_no >= node->get_size()) {
        throw IndexEntryNotFoundError();
    }
    // buffer_pool_manager_->unpin_page(node->get_page_id(), false);  // unpin it!
    Rid rid = *node->get_rid(iid.slot_no);
    delete node;
    return rid;
}

/**
 * @brief find_leaf_page + lower_bound，找到第一个大于等于key的索引槽
 *
 * @param key
 * @return Iid
 * @note 上层传入的key本来是int类型，通过(const char *)&key进行了转换
 * 可用*(int *)key转换回去
 */
Iid IxIndexHandle::lower_bound(const char *key) {
    auto pair = find_leaf_page(key, Operation::FIND, nullptr);
    IxNodeHandle *node = pair.first;
    int key_idx = node->lower_bound(key);

    Iid iid = {.page_no = node->get_page_no(), .slot_no = key_idx};

    // unpin leaf node
    unlock_pages(nullptr, Operation::FIND, pair.second);
    // buffer_pool_manager_->unpin_page(node->get_page_id(), false);
    delete node;
    return iid;
}

/**
 * @brief find_leaf_page + upper_bound， 找到第一个大于key的索引槽
 *
 * @param key
 * @return Iid
 */
Iid IxIndexHandle::upper_bound(const char *key) {
    auto pair = find_leaf_page(key, Operation::FIND, nullptr);
    IxNodeHandle *node = pair.first;
    int key_idx = node->upper_bound(key);

    Iid iid;
    if (key_idx == node->get_size()) {
        // 这种情况无法根据iid找到rid，即后续无法调用ih->get_rid(iid)
        iid = leaf_end();
    } else {
        iid = {.page_no = node->get_page_no(), .slot_no = key_idx};
    }

    // unpin leaf node
    unlock_pages(nullptr, Operation::FIND, pair.second);
    // buffer_pool_manager_->unpin_page(node->get_page_id(), false);
    delete node;
    return iid;
}

/**
 * @brief 指向最后一个叶子的最后一个结点的后一个
 * 用处在于可以作为IxScan的最后一个
 *
 * @return Iid
 */
Iid IxIndexHandle::leaf_end() const {
    IxNodeHandle *node = fetch_node(file_hdr_->last_leaf_);
    Iid iid = {.page_no = file_hdr_->last_leaf_, .slot_no = node->get_size()};
    // buffer_pool_manager_->unpin_page(node->get_page_id(), false);  // unpin it!
    delete node;
    return iid;
}

/**
 * @brief 指向第一个叶子的第一个结点
 * 用处在于可以作为IxScan的第一个
 *
 * @return Iid
 */
Iid IxIndexHandle::leaf_begin() const {
    Iid iid = {.page_no = file_hdr_->first_leaf_, .slot_no = 0};
    return iid;
}

/**
 * @brief 获取一个指定结点
 *
 * @param page_no
 * @return IxNodeHandle*
 * @note pin the page, remember to unpin it outside!
 */
IxNodeHandle *IxIndexHandle::fetch_node(int page_no) const {
    Page *page = buffer_pool_manager_->fetch_page(PageId{fd_, page_no});
    IxNodeHandle *node = new IxNodeHandle(file_hdr_, page);

    return node;
}

/**
 * @brief 创建一个新结点
 *
 * @return IxNodeHandle*
 * @note pin the page, remember to unpin it outside!
 * 注意：对于Index的处理是，删除某个页面后，认为该被删除的页面是free_page
 * 而first_free_page实际上就是最新被删除的页面，初始为IX_NO_PAGE
 * 在最开始插入时，一直是create node，那么first_page_no一直没变，一直是IX_NO_PAGE
 * 与Record的处理不同，Record将未插入满的记录页认为是free_page
 */
IxNodeHandle *IxIndexHandle::create_node() {
    IxNodeHandle *node;
    file_hdr_->num_pages_++;

    PageId new_page_id = {.fd = fd_, .page_no = INVALID_PAGE_ID};
    // 从3开始分配page_no，第一次分配之后，new_page_id.page_no=3，file_hdr_.num_pages=4
    Page *page = buffer_pool_manager_->new_page(&new_page_id);
    node = new IxNodeHandle(file_hdr_, page);
    return node;
}

/**
 * @brief 从node开始更新其父节点的第一个key，一直向上更新直到根节点
 *
 * @param node
 */
void IxIndexHandle::maintain_parent(IxNodeHandle *node) {
    IxNodeHandle *curr = node;
    while (curr->get_parent_page_no() != IX_NO_PAGE) {
        // Load its parent
        IxNodeHandle *parent = fetch_node(curr->get_parent_page_no());
        int rank = parent->find_child(curr);
        char *parent_key = parent->get_key(rank);
        char *child_first_key = curr->get_key(0);
        if (memcmp(parent_key, child_first_key, file_hdr_->col_tot_len_) == 0) {
            // assert(buffer_pool_manager_->unpin_page(parent->get_page_id(), true));
            break;
        }
        memcpy(parent_key, child_first_key, file_hdr_->col_tot_len_);  // 修改了parent node
        if(curr != node) delete curr;
        curr = parent;

        // assert(buffer_pool_manager_->unpin_page(parent->get_page_id(), true));
    }
    if (curr != node) delete curr;
}

/**
 * @brief 要删除leaf之前调用此函数，更新leaf前驱结点的next指针和后继结点的prev指针
 *
 * @param leaf 要删除的leaf
 */
void IxIndexHandle::erase_leaf(IxNodeHandle *leaf) {
    assert(leaf->is_leaf_page());

    IxNodeHandle *prev = fetch_node(leaf->get_prev_leaf());
    prev->set_next_leaf(leaf->get_next_leaf());
    // buffer_pool_manager_->unpin_page(prev->get_page_id(), true);

    IxNodeHandle *next = fetch_node(leaf->get_next_leaf());
    next->set_prev_leaf(leaf->get_prev_leaf());  // 注意此处是SetPrevLeaf()
    // buffer_pool_manager_->unpin_page(next->get_page_id(), true);

    delete prev;
    delete next;
}

/**
 * @brief 删除node时，更新file_hdr_.num_pages
 *
 * @param node
 */
void IxIndexHandle::release_node_handle(IxNodeHandle &node) { file_hdr_->num_pages_--; }

/**
 * @brief 将node的第child_idx个孩子结点的父节点置为node
 */
void IxIndexHandle::maintain_child(IxNodeHandle *node, int child_idx) {
    if (!node->is_leaf_page()) {
        //  Current node is inner node, load its child and set its parent to current node
        int child_page_no = node->value_at(child_idx);
        IxNodeHandle *child = fetch_node(child_page_no);
        child->set_parent_page_no(node->get_page_no());
        // buffer_pool_manager_->unpin_page(child->get_page_id(), true);
        delete child;
    }
}

/**
 * @brief 判断一个节点是否是 "safe"
 * 结点插入一个键值对后仍然未满(size+1<max_size)；
 * 或者结点删除一个键值对后仍然超过或等于半满(size-1>=min_size)；
 * 注意，根结点的min_size=2，其余结点的min_size=max_size/2。
 *
 * @param node 需要判断的节点
 * @param operation 操作类型
 * @return true 节点是一个 "safe" 节点
 * @return false 节点不是 "safe" 节点
 */
bool IxIndexHandle::is_node_safe(IxNodeHandle *node, Operation operation) const {
    if (operation == Operation::FIND) {
        return true;
    }
    int min_sz = node->is_root_page() ? 2 : node->get_min_size();
    if (operation == Operation::DELETE) {
        return node->get_size() - 1 >= min_sz;
    }
    if (operation == Operation::INSERT) {
        return node->get_size() + 1 < node->get_max_size();
    }
    return false;
}

/**
 * @brief 释放 transaction 中 pageset 的锁 以及根节点的锁
 *
 * @param transaction
 * @param operation
 * @param is_root_lock 表示根节点是否被上锁
 */
void IxIndexHandle::unlock_pages(Transaction *transaction, Operation operation, bool &is_root_lock) {
    if (is_root_lock) {
        is_root_lock = false;
        root_latch_.unlock();
    }
    if(transaction != nullptr){
        auto page_set = transaction->get_index_latch_page_set().get();
        while (!page_set->empty()) {
            page_set->pop_front();
        }
    }
}

void IxIndexHandle::delete_pages(Transaction *transaction) {
    if (transaction) {
        auto deleted_page_set = transaction->get_index_deleted_page_set().get();
        while (!deleted_page_set->empty()) {
            buffer_pool_manager_->delete_page(deleted_page_set->front()->get_page_id());
            deleted_page_set->pop_front();
        }
    }
}
