/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <algorithm>
#include <cfloat>
#include <climits>
#include <functional>
#include <iostream>
#include <map>
#include <memory>
#include <string>
#include <vector>
#include "defs.h"
#include "errors.h"
#include "sm_defs.h"
#include "common/common.h"

/* 字段元数据 */
struct ColMeta {
    std::string tab_name;   // 字段所属表名称
    std::string name;       // 字段名称
    ColType type;           // 字段类型
    int len;                // 字段长度
    int offset;             // 字段位于记录中的偏移量
    bool index;             
    ast::GroupByCond group_by_cond = ast::NONE;
    std::string alias;

    friend std::ostream &operator<<(std::ostream &os, const ColMeta &col) {
        // ColMeta中有各个基本类型的变量，然后调用重载的这些变量的操作符<<（具体实现逻辑在defs.h）
        return os << col.tab_name << ' ' << col.name << ' ' << col.type << ' ' << col.len << ' ' << col.offset << ' '
                  << col.index;
    }

    friend std::istream &operator>>(std::istream &is, ColMeta &col) {
        return is >> col.tab_name >> col.name >> col.type >> col.len >> col.offset >> col.index;
    }

    static ColMeta createCountAllColMeta(const TabCol& tableColumn) {
        return { "", "*", TYPE_INT, sizeof(int), 0, false, ast::COND_COUNT, tableColumn.alias };
    }
};

/* 索引元数据 */
struct IndexMeta {
    std::string tab_name;           // 索引所属表名称
    int col_tot_len;                // 索引字段长度总和
    int col_num;                    // 索引字段数量
    std::vector<ColMeta> cols;      // 索引包含的字段
    std::string index_name;         // 索引名称

    friend std::ostream &operator<<(std::ostream &os, const IndexMeta &index) {
        os << index.tab_name << " " << index.col_tot_len << " " << index.col_num;
        for(auto& col: index.cols) {
            os << "\n" << col;
        }
        return os;
    }

    friend std::istream &operator>>(std::istream &is, IndexMeta &index) {
        is >> index.tab_name >> index.col_tot_len >> index.col_num;
        for(int i = 0; i < index.col_num; ++i) {
            ColMeta col;
            is >> col;
            index.cols.push_back(col);
        }
        return is;
    }

    std::unique_ptr<char[]> get_key(const char *buf) const{
        std::unique_ptr<char[]> key(new char[col_tot_len]);
        int offset = 0;
        for(auto& col:cols){
            memcpy(key.get()+offset,buf+col.offset,col.len);
            offset += col.len;
        }
        return key;
    }

    std::unique_ptr<char[]> constructKey(const std::function<void(char*, const ColMeta&)>& setValue) {
        std::unique_ptr<char[]> key(new char[col_tot_len]);
        int offset = 0;
        for(const auto& col: cols) {
            switch(col.type) {
                case ColType::TYPE_INT:
                    *(int*)(key.get() + offset) = std::numeric_limits<int>::min();
                    break;
                case ColType::TYPE_FLOAT:
                    *(double*)(key.get() + offset) = std::numeric_limits<double>::min();
                    break;
                case ColType::TYPE_STRING:
                    memset(key.get() + offset, '\0', col.len);
                    break;
                default:
                    throw RMDBError("Unknown type");
            }
            offset += col.len;
        }
        setValue(key.get(), cols[0]);
        return key;
    }

    std::unique_ptr<char[]> get_min_key() {
        return constructKey([](char *key, const ColMeta& meta){});
    }

    std::unique_ptr<char[]> get_max_key() {
        return constructKey([](char *key, const ColMeta& meta){
            switch(meta.type) {
                case ColType::TYPE_INT:
                    *(int*)(key) = std::numeric_limits<int>::max();
                    break;
                case ColType::TYPE_FLOAT:
                    *(double*)(key) = std::numeric_limits<double>::max();
                    break;
                case ColType::TYPE_STRING:
                    memset(key, '\x7f', meta.len);
                    break;
                default:
                    throw RMDBError("Unknown type");
            }
        });
    }

    void set_key(char *key, const std::string &col_name, const Value &val){
        int offset = 0;
        size_t i = 0;
        for (; i < cols.size(); i ++) {
            if (cols[i].name == col_name) {
                break;
            }
            offset += cols[i].len;
        }
        auto &col = cols[i];
        if (col.type == val.type) {
            memcpy(key + offset, val.raw->data, col.len);
            return;
        }
        std::unique_ptr<char[]> buf(new char[col.len]);
        // 只涉及 INT, BIGINT, FLOAT 三种类型 之间的转换
        switch(col.type) {
            case ColType::TYPE_INT:
                if (val.type == ColType::TYPE_FLOAT)
                    *(int*)(buf.get()) = static_cast<int>(*(double *)val.raw->data);
                break;
            case ColType::TYPE_FLOAT:
                if (val.type == ColType::TYPE_INT)
                    *(double*)(buf.get()) = static_cast<double>(*(int *)val.raw->data);
                break;
            default:
                throw RMDBError("Unknown type");
        }
        memcpy(key + offset, buf.get(), col.len);
    }
};

inline Value convertValue(char *base, const ColMeta& meta) {
    Value value;
    switch (meta.type) {
        case TYPE_INT:
            value.set_int(*(int *) (base + meta.offset));
            break;
        case TYPE_FLOAT:
            value.set_float(*(float *) (base + meta.offset));
            break;
        case TYPE_STRING: {
            std::string str((char *) (base + meta.offset), meta.len);
            str.resize(std::min(str.find('\0'), (size_t)meta.len));
            value.set_str(str);
            break;
        }
        default:
            throw InternalError("not implemented");
    }
    return value;
}

/* 表元数据 */
struct TabMeta {
    std::string name;                   // 表名称
    std::vector<ColMeta> cols;          // 表包含的字段
    std::vector<IndexMeta> indexes;     // 表上建立的索引

    TabMeta(){}

    TabMeta(const TabMeta &other) {
        name = other.name;
        for(auto col : other.cols) cols.push_back(col);
    }

    /**
     * @brief 在 TabMeta 中删除索引
     * 
     * @param col_names 
     */
    void drop_index(const std::vector<std::string>& col_names){
        for(auto index_iter = indexes.begin(); index_iter != indexes.end(); ++index_iter) {
            if(index_iter->col_num == (int)col_names.size()) {
                int i = 0;
                for(; i < index_iter->col_num; ++i) {
                    if(index_iter->cols[i].name.compare(col_names[i]) != 0)
                        break;
                }
                if(i == index_iter->col_num) {
                    indexes.erase(index_iter);
                    return;
                }
            }
        }
    }

    /* 判断当前表中是否存在名为col_name的字段 */
    bool is_col(const std::string &col_name) const {
        auto pos = std::find_if(cols.begin(), cols.end(), [&](const ColMeta &col) { return col.name == col_name; });
        return pos != cols.end();
    }

    /* 判断当前表上是否建有指定索引，索引包含的字段为col_names */
    bool is_index(const std::vector<std::string>& col_names) const {
        for(auto& index: indexes) {
            if(index.col_num == col_names.size()) {
                size_t i = 0;
                for(; i < index.col_num; ++i) {
                    if(index.cols[i].name.compare(col_names[i]) != 0)
                        break;
                }
                if(i == index.col_num) return true;
            }
        }

        return false;
    }

    /* 根据字段名称集合获取索引元数据 */
    std::vector<IndexMeta>::iterator get_index_meta(const std::vector<std::string>& col_names) {
        for(auto index = indexes.begin(); index != indexes.end(); ++index) {
            if((*index).col_num != col_names.size()) continue;
            auto& index_cols = (*index).cols;
            size_t i = 0;
            for(; i < col_names.size(); ++i) {
                if(index_cols[i].name.compare(col_names[i]) != 0) 
                    break;
            }
            if(i == col_names.size()) return index;
        }
        throw IndexNotFoundError(name, col_names);
    }

    /* 根据字段名称获取字段元数据 */
    std::vector<ColMeta>::iterator get_col(const std::string &col_name) {
        auto pos = std::find_if(cols.begin(), cols.end(), [&](const ColMeta &col) { return col.name == col_name; });
        if (pos == cols.end()) {
            throw ColumnNotFoundError(col_name);
        }
        return pos;
    }

    friend std::ostream &operator<<(std::ostream &os, const TabMeta &tab) {
        os << tab.name << '\n' << tab.cols.size() << '\n';
        for (auto &col : tab.cols) {
            os << col << '\n';  // col是ColMeta类型，然后调用重载的ColMeta的操作符<<
        }
        os << tab.indexes.size() << "\n";
        for (auto &index : tab.indexes) {
            os << index << "\n";
        }
        return os;
    }

    friend std::istream &operator>>(std::istream &is, TabMeta &tab) {
        size_t n;
        is >> tab.name >> n;
        for (size_t i = 0; i < n; i++) {
            ColMeta col;
            is >> col;
            tab.cols.push_back(col);
        }
        is >> n;
        for(size_t i = 0; i < n; ++i) {
            IndexMeta index;
            is >> index;
            tab.indexes.push_back(index);
        }
        return is;
    }
};

// 注意重载了操作符 << 和 >>，这需要更底层同样重载TabMeta、ColMeta的操作符 << 和 >>
/* 数据库元数据 */
class DbMeta {
    friend class SmManager;

   private:
    std::string name_;                      // 数据库名称
    std::map<std::string, TabMeta> tabs_;   // 数据库中包含的表

   public:
    // DbMeta(std::string name) : name_(name) {}

    /* 判断数据库中是否存在指定名称的表 */
    bool is_table(const std::string &tab_name) const { return tabs_.find(tab_name) != tabs_.end(); }

    void SetTabMeta(const std::string &tab_name, const TabMeta &meta) {
        tabs_[tab_name] = meta;
    }

    /* 获取指定名称表的元数据 */
    TabMeta &get_table(const std::string &tab_name) {
        auto pos = tabs_.find(tab_name);
        if (pos == tabs_.end()) {
            throw TableNotFoundError(tab_name);
        }

        return pos->second;
    }

    // 重载操作符 <<
    friend std::ostream &operator<<(std::ostream &os, const DbMeta &db_meta) {
        os << db_meta.name_ << '\n' << db_meta.tabs_.size() << '\n';
        for (auto &entry : db_meta.tabs_) {
            os << entry.second << '\n';
        }
        return os;
    }

    friend std::istream &operator>>(std::istream &is, DbMeta &db_meta) {
        size_t n;
        is >> db_meta.name_ >> n;
        for (size_t i = 0; i < n; i++) {
            TabMeta tab;
            is >> tab;
            db_meta.tabs_[tab.name] = tab;
        }
        return is;
    }
};
